
import math

###### DEPRECIATION ########
# Vector is being depreciated in favor of using a tuple(int, int) like Pyglet's sprites.
# Stop using Vector immediately and remove references to Vector as soon as possible
#TODO remove Vector
# WARNING

class Vector:
   """
   :param x: A position in space on the horizontal plane
   :param y: A position in space on the vertical plane
   :type x: int
   :type y: int

   A typical Vector datatype
   """
   def __init__(self, x, y):
      self.x = x
      self.y = y

   def __add__(self, other):
      """
      :param other: A second vector
      :param type: Vector()

      Add two vectors together. Changes the first vector
      """
      self.x += other.x
      self.y += other.y

   def __sub__(self, other):
      """
      :param other: A second vector
      :param type: Vector()

      Substract two vectors from each other. Changes the first vector
      """
      self.x -= other.x
      self.y -= other.y

   def __mul__(self, other):
      """
      
      Multiples a vector by a scalar value
      other is a number
      """
      self.x *= other
      self.y *= other

   def __truediv__(self, other):
      """
      Divides a vector by a scalar value
      other is a number
      """
      self.x /= other
      self.y /= other

   def __floordiv__(self, other): #Scalar
      """
      Floor division by a scalar value
      other is a number
      """
      self.x = self.x // other
      self.y = self.y // other

   def __mod__(self, other): #Scalar
      """
      Modulus by a scalar value
      other is a number
      """
      self.x = self.x % other
      self.y = self.y % other

   def __pow__(self, other): #Scalar
      """
      Exponent by a scalar value
      other is a number
      """
      self.x = self.x ** other
      self.y = self.y ** other

   def __eq__(self, other):
      """
      :param other: A second vector
      :param type: Vector()
      :return: True if x and y are the same for both vectors, False if not
      :rtype: bool

      Equality between two vectors
      """
      if self.x == other.x and self.y == other.y:
         return True
      else:
         return False

   def __ne__(self, other):
      if self.x != other.x and self.y != other.y:
         return True
      else:
         return False

   def __lt__(self, other):
      if self.x < other.x and self.y < other.y:
         return True
      else:
         return False

   def __gt__(self, other):
      if self.x > other.x and self.y > other.y:
         return True
      else:
         return False

   def __le__(self, other):
      if self.x <= other.x and self.y <= other.y:
         return True
      else:
         return False

   def __ge__(self, other):
      if self.x >= other.x and self.y >= other.y:
         return True
      else:
         return False

   def dot(self, other):
      """
      :param other: A second vector
      :param type: Vector()
      :return: Dot product of two vectors
      :rtype: float

      Dot product of two vectors.
      """
      return self.x * other.x + self.y * other.y

   def cross(self, other):
      """
      :param other: A second vector
      :param type: Vector()
      :return: Cross product of two vectors
      :rtype: float

      Cross product of two vectors.
      """
      return self.x * other.y - self.y * other.x

   def mag_square(self):
      """
      :return: Square magnitude of the vector.
      :rtype: float

      Square magnitude of the vector.
      """
      return self.x * self.x + self.y * self.y

   def mag(self):
      """
      :return: Magnitude of the vector.
      :rtype: float
      
      Magnitude of the vector.
      """
      return math.sqrt(self.mag_square())

   def angle_self(self):
      """
      :return: Angle of the vector.
      :rtype: float

      Angle of the vector.
      returns a number.
      """
      return math.atan2(self.x, self.y)

   def normal(self):
      """
      :return: Normal of the vector.
      :rtype: float
      
      Normal of the vector.
      """
      len = self.mag_square()
      if len != 0:
         len = math.sqrt(len)
         self.x = self.x // len
         self.y = self.y // len
      else:
         self.x = 0
         self.y = 0

   def dist_square(self, x, y):
      """
      :param x: A position in space
      :param y: A position in space
      :type x: int
      :type y: int
      :return: Distance squared from x, y.
      :rtype: float
      
      Distance squared from x, y. Faster than dist() for relative distance comparisons
      """
      xd = self.x - x
      yd = self.y - y
      return xd * xd + yd * yd

   def dist(self, x, y):
      """
      :param x: A position in space
      :param y: A position in space
      :type x: int
      :type y: int
      :return: Distance squared from x, y.
      :rtype: float
      
      Distance from x, y
      """
      return math.sqrt(self.dist_square(x, y))

   def angle_rads(self, other):
      """
      :param other: Second vector to find the angle to
      :type other: Vector()
      :return: Angle to the second vector in radians
      :rtype: float

      Angle to a vector in radians
      """
      return math.atan2(self.cross(other), self.dot(other))

   def angle(self, other):
      """
      :param other: Second vector to find the angle to
      :type other: Vector()
      :return: Angle to the second vector in degrees
      :rtype: float

      Angle to a vector in degrees
      """
      return math.degrees(math.atan2(self.cross(other), self.dot(other)))

   def angle_point(self, other):
      """
      :param other: Second vector to find the angle to
      :type other: Vector()
      :return: Angle point to the second vector
      :rtype: float

      Angle point to a vector
      """
      return math.atan2(self.x - other.y, self.x - other.x)

   def abs(self):
      """
      Absolute value.
      Sets x and y to positive values.
      """
      self.x = abs(self.x)
      self.y = abs(self.y)

   def project(self, other):
      """
      :param other: Second vector to project from
      :type other: Vector()
      :return: Value of the projection
      :rtype: float

      Projection from another vector
      """
      return other * (self.dot(other) / other.mag_square())

   def clamp(self, min, max):
      """
      :param min: Minimum value of x and y
      :param max: Maximum values of x and y
      :type min: int
      :type max: int

      Sets the vector to a minimum and maximum value
      """
      if self.x < min:
         self.x = min
      elif self. x > max:
         self.x = max

      if self.y < min:
         self.y = min
      elif self.y > max:
         self.y = max

   def move(other, delta):
      """
      :param other: Direction to move in
      :param delta: Units of time to move in
      :type other: Vector()
      :type delta: float

      Move in the direction of other by delta units
      """
      self.x += other.x * delta
      self.y += other.y * delta 

   def __str__(self):
      """
      :return: String representation of the vector
      :rtype: string
      
      String representation of a vector
      """
      return "{int(self.x)}:{int(self.y)}"





import ctypes
import time


class Clock:
   """
   :param rate: Minimum amount of time before the next tick can occur
   :type rate: float

   Keeps track of time through polling tick
   """

   def __init__(self, rate):
      self.rate = rate #: float, Minimum amount of time before the next tick can occur
      self.time = ctypes.c_ulong(0) #: ctypes.c_ulong, Number of ticks. Using ctypes for free rollover 
      self.last = time.time() #: time.time, Time of the last tick

   def get_time(self):
      """
      Get the number of ticks that have elapsed
      :return: Number of ticks that have elapsed
      :rtype: int
      """
      return self.time.value

   def tick(self):
      """
      :return: delta or None
      :rtype: float or None
   
      Attemps to advance the system clock. 
      Delta is the ratio of time that has passed since the tick. 
      For example if your tick rate is 100ms, and your tick happens 150ms after your last tick, your delta will be 1.5
      You probably are going to want to use this one for timing in games
      """

      now = time.time()
      value = now - self.last
      if value > self.rate:
         delta = value / self.rate
         self.time.value += 1
         self.last = now
         return delta
      else:
         return None

   def tick_time(self):
      """
      :return: Current time
      :rtype: int

      Attemps to advance the system clock.
      Always returns the time.
      """
      now = time.time()
      value = now - self.last
      if value > self.rate:
         self.last = now
         self.time.value += 1
      return value

   def tick_check(self):
      """
      :return: True if rate time has elapsed since the last tick, False if not
      :rtype: bool

      Attemps to advance the system clock. 
      """
      now = time.time()
      if now - self.last > self.rate:
         self.last = now
         self.time.value += 1
         return True
      else:
         return False


class SyncClock:
   """
   A Clock for syncing values from a master Clock.
   """

   def __init__(self):
      self.time = 0

   def get_time(self):
      """
      :return: The current time
      :rtype: int

      Gets the current time
      """
      return self.time




class Shape:
   """
   :param x: Position of the shape in the x-axis
   :param y: Position of the shape in the y-axis
   :type x: Int
   :type y: Int

   Abstract Base Shape that other shapes extend from. 
   Shapes are simple data structures that describe the physical properties of an body to a Collider.
   """
   def __init__(self, x, y):
      self.position = (x, y) #: Tuple(Int, Int), position of the shape


class Rect(Shape):
   """
   :param x: Position of the shape in the x-axis
   :param y: Position of the shape in the y-axis
   :param width: Width of the Rect
   :param height: Height of the Rect
   :type x: Int
   :type y: Int
   :type width: int
   :type height: int

   A Rectangle
   Treated as Axis-Aligned Bounding Boxes by the Collider.
   """
   def __init__(self, x, y, width, height):
      self.position = (x, y) #: Tuple(Int, Int), position of the shape
      self.width = width #: Int, Width, extending up
      self.height = height #: Int, Height, extending to the right


class Circle(Shape):
   """
   :param x: Position of the shape in the x-axis
   :param y: Position of the shape in the y-axis
   :param radius: Width of the Rect
   :type x: Int
   :type y: Int
   :type radius: int

   A Circle
   """
   def __init__(self, x, y, radius):
      self.position = (x, y) #: Tuple(Int, Int), position of the shape
      self.radius = radius #: Int, Radius, extending from the center.






from pecrs.shape import *


class Collider:
   """
   Collider handles collisions between two objects with physical dimensions.
   Note that the collider is abstract and low-level, dealing with any set of positions and dimensions.
   """

   def check(self, shape, other):
      """
      :param shape: Shape of the first object
      :param other: Shape of the second object
      :type shape: Shape
      :type other: Shape
      :return: True if there is a collision, False is not
      :rtype: bool

      Checks for a collision between any two shapes. Determines what the shapes are and sends them to the appropriate collision method.
      If you already know the shape of your objects, use a more direct method.
      """
      if type(shape) == Rect:
         if type(other) == Rect:
            return self.rect_rect(shape.width, shape.height, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1])
         elif type(other) == Circle:
            return self.rect_circle(shape.width, shape.height, shape.position[0], shape.position[1], other.radius, other.position[0], other.position[1])
      elif type(shape) == Circle:
         if type(other) == Rect:
            return self.circle_rect(shape.radius, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1])
         elif type(other) == Circle:
            return self.circle_circle(shape.radius, shape.position[0], shape.position[1], other.radius, other.position[0], other.position[1])

   def check_rects(self, shape, other):
      """
      :param shape: Shape of the first Rect
      :param other: Shape of the second Rect
      :type shape: Rect
      :type other: Rect
      :return: True if there is a collision, False is not
      :rtype: bool

      Checks for a collision between any two Rects. This is a higher level interface for Collider.rect_rect().
      Note that Pyglet sprites are effectively Rects and can be used interchangably here.
      """
      return self.rect_rect(shape.width, shape.height, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1])

   def check_circles(self, shape, other):
      """
      :param shape: Shape of the first Circle
      :param other: Shape of the second Circle
      :type shape: Circle
      :type other: Circle
      :return: True if there is a collision, False is not
      :rtype: bool

      Checks for a collision between any two Circles. This is a higher level interface for Collider.circle_circle().
      """
      return self.circle_circle(shape.radius, shape.position[0], shape.position[1], other.radius, other.position[0], other.position[1])

   def rect_rect(self, w, h, x, y, wo, ho, xo, yo):
      """
      :param w: Width of the first Rect
      :param h: Height of the first Rect
      :param x: x position of the first Rect
      :param y: y position of the first Rect
      :param wo: Width of the second Rect
      :param ho: Height of the second Rect
      :param xo: x position of the second Rect
      :param yo: y position of the second Rect
      :type w: Int
      :type h: Int
      :type x: Int
      :type y: Int
      :type wo: Int
      :type ho: Int
      :type xo: Int
      :type yo: Int

      Checks for a collision between two Rects in abstract.
      """
      if x < xo + wo and x + w > xo and y < yo + ho and y + h > yo:
         return True
      else:
         return False

   def rect_circle(self, w, h, x, y, ro, xo, yo):
      """
      :param w: Width of the first Rect
      :param h: Height of the first Rect
      :param x: x position of the first Rect
      :param y: y position of the first Rect
      :param ro: Radius of the second Circle
      :param xo: x position of the second Circle
      :param yo: y position of the second Circle
      :type w: Int
      :type h: Int
      :type x: Int
      :type y: Int
      :type ro: Int
      :type xo: Int
      :type yo: Int

      Checks for a collision between a Rect and a Circle in abstract.
      """
      if xo < x:
         x = x
      elif xo > x + w:
         x = x + w
      else:
         x = xo

      if yo < y:
         y = y
      elif yo > y + h:
         y = y + h
      else:
         y = yo

      dx = x - xo
      dy = y - yo
      if dx * dx + dy * dy < ro * ro:
         return True
      else:
         return False

   def circle_rect(self, r, x, y, wo, ho, xo, yo):
      """
      :param r: Radius of the first Circle
      :param x: x position of the first Circle
      :param y: y position of the first Circle
      :param wo: Width of the second Rect
      :param ho: Height of the second Rect
      :param xo: x position of the second Rect
      :param yo: y position of the second Rect
      :type r: Int
      :type x: Int
      :type y: Int
      :type wo: Int
      :type ho: Int
      :type xo: Int
      :type yo: Int

      Checks for a collision between a Circle and a Rect in abstract. Interface for Collider.rect_circle
      """
      return self.rect_circle(wo, ho, xo, yo, r, x, y)

   def circle_circle(self, r, x, y, ro, xo, yo):
      """
      :param r: Radius of the first Circle
      :param x: x position of the first Circle
      :param y: y position of the first Circle
      :param ro: Radius of the second Circle
      :param xo: x position of the second Circle
      :param yo: y position of the second Circle
      :type r: Int
      :type x: Int
      :type y: Int
      :type wo: Int
      :type ho: Int
      :type xo: Int
      :type yo: Int

      Checks for a collision between any two Circles abstract.
      """
      dx = x - xo
      dy = y - yo
      rs = r + ro
      if dx * dx + dy * dy < rs * rs:
         return True
      else:
         return False






class SpatialHash:
  """
  :param size: Size of each grid cell
  :type size: int

  A generic spatialhash
  Keeps a bucket of objects at a given scaled position
  """
  def __init__(self, size):
    self.grid = {} #: dict, Grid of buckets
    self.size = size #: int, Size of each grid cell
  
  def scale(self, x, y):
    """
    :param x: A position in to be scaled on the horizontal plane
    :param y: A position in to be scaled on the horizontal plane
    :type x: int
    :type y: int
    :return: Grid cell key for the position at x, y
    :rtype: tuple(int, int)

    Scales an x, y value to a grid cell position
    """
    return (x // self.size, y // self.size)

  def add(self, item, area):
    """
    :param item: Item to be added to the grid
    :param area: Area to add the item to
    :type item: Object
    :type area: tuple(int, int)

    Add an item to the grid at an area
    """

    if area not in self.grid:
      self.grid[area] = []
    self.grid[area].append(item)

  def add_pos(self, item, x, y):
    """
    :param item: Item to be added to the grid
    :param x: Area position
    :param y: Area position
    :type item: Object
    :type x: int
    :type y: int

    Add an item to the grid at x, y
    """
    return self.add(item, self.scale(x, y))

  def has(self, item, area=None):
    """
    :param item: Item to be added to the grid
    :param area: Area to check, checks all if None
    :type item: Object
    :type area: tuple(int, int) or None

    Checks to see if an item is in the grid
    """
    if area:
      if area in self.grid:
        if item in self.grid[area]:
          return True

    else:
      for area in self.grid:
        if item in self.grid[area]:
          return True

    return False

  def get(self, area):
    """
    :param area: Grid area
    :type area: tuple(int, int)
    :return: List of objects at position at x, y
    :rtype: list(Object)

    Get a bucket at area from the grid
    """
    if area not in self.grid:
      self.grid[area] = []
    return self.grid[area]

  def get_pos(self, x, y):
    """
    :param x: Area position
    :param y: Area position
    :type x: int
    :type y: int
    :return: List of objects at position at x, y
    :rtype: list(Object)

    Get a bucket at x, y from the grid
    """
    return self.get(self.scale(x, y))
  
  def delete(self, item, area):
    """
    :param item: Item to be removed to the grid
    :param area: Area to remove the item from
    :type item: Object
    :type area: tuple(int, int)

    Remove an item from the grid at an area
    """
    if area in self.grid:
      if item in self.grid[area]:
        self.grid[area].remove(item)
        if len(self.grid[area]) == 0:
          del self.grid[area]

  def delete_pos(self, item, x, y):
    """
    :param item: Item to be removed to the grid
    :param x: Area position
    :param y: Area position
    :type item: Object
    :type x: int
    :type y: int

    Remove an item to the grid at x, y
    """
    return self.delete(item, self.scale(x, y))

      




class Index:
   """
   Indexing system for assigning unique IDs to bodies.

   Index keeps track what number to assign to new entities.

   TODO move object list keeping and retrival; update docs
   """

   def __init__(self):
      self.count = -1 #: The ID to be used for the next new body
      self.free = [] #: A list of free IDs for reuse
      
   def next(self):
      """
      :return: A unique id
      :rtype: int

      Gets a unique identifer for this index. Freed ids are assigned first, in first-in-last-out order.
      """
      if not self.free:
         self.count += 1
         return self.count
      else:
         return self.free.pop()

   def delete(self, id):
      """
      :param id: Previously assigned id to be returned to the system
      :type id: int

      Frees an identifer to be reused.
      """
      if id <= self.count:         
         self.free.append(id)
      




from pecrs.spatialhash import SpatialHash
from pecrs.collider import Collider
from pecrs.index import Index

class Space:
   """
   :param size: Size of the collision areas
   :type size: int

   Space handles the movement and collision detection between shapes.
   You can change the values here to work with different spatial partioning system and collision systems.
   """

   def __init__(self, size=256):
      self.grid = SpatialHash(size) #: SpatialHash, spatial partitioning system. Think about combining these into one class
      self.collider = Collider() #: Collider, collision detection system
      self.index = Index()
      self.ids = {}
      self.shapes = {}
      self.areas = {}
      self.moving = []
      self.directions = {}
      self.actives = []
      self.groups = {}
      self.collidable = {}

   def add(self, shape, id=-1, collidable=True, active=True, moving=False, group=None):
      """
      :param shape: Shape to add to the space
      :type shape: Shape
      
      Adds a shape to the space.
      """
      if id < 0:
         id = self.index.next()
      self.ids[shape] = id
      self.shapes[id] = shape
      area = self.grid.scale(shape.position[0], shape.position[1])   
      self.areas[shape] = area
      self.add_to_grid(shape, area)
      self.collidable[shape] = collidable

      if moving:
         self.moving.append(shape)
      
      if active:
         self.actives.append(shape)
         self.directions[shape] = (0, 0)
      
      if group:
         if group in self.groups:
            self.groups[group].append(shape)
         else:
            self.groups[group] = [shape]

   def delete(self, shape, group=None):
      """
      :param shape: Shape to removed from the space
      :type shape: Shape

      Removes a shape from the space.
      """
      if shape in self.ids:
         id = self.ids[shape]
         del self.ids[shape]
         del self.shapes[id]
         del self.moving[shape]
         self.delete_from_grid(shape, self.areas[shape])
         del self.areas[shape]
         if shape in self.actives: #TODO I think self.actives, self.directions might be work better with groups if given?
            self.actives.remove(shape)
         if shape in self.directions:
            del self.directions[shape]
         if group:
            self.groups[group].remove(shape)

   def delete_id(self, id, group=None):
      shape = self.ids[shape]
      self.delete(shape, group)

   def delete_group(self, group):
      for shape in self.groups[group]:
         self.delete(shape)
      del self.groups[group]
      
   def has(self, shape):
      """
      :param shape: Shape to check for
      :type shape: Shape
      :return: True if the space contains the shape, False if not
      :rtype: bool

      Checks if a space has a shape in that shape's area
      Note that the shape's must be set correctly or errors can occur
      """
      return shape in self.ids

   def has_id(self, id):
      return id in self.shapes

   def collisions_at(self, x, y, width=1, height=1):
      """
      :param x: Posistion to search for on the x-axis
      :param y: Posistion to search for on the y-axis
      :param width: Width of the search area
      :param height: Height of the search area
      :type x: Int
      :type y: Int
      :type width: Int
      :type height: Int
      :return: List of shapes colliding at the position
      :rtype: list(Shapes)

      Gets a list of all shapes colliding at x, y
      """
      collisions = []
      for other in self.grid.get(self.grid.scale(x, y)):
         if self.collidable[other]:
            if self.collider.rect_rect(width, height, x, y, other.width, other.height, other.position[0], other.position[1]):
               collisions.append(other)
      return collisions

   def collisions_with(self, shape):
      """
      :param shape: Shape to find collisions with
      :type shape: Shape
      :return: List of shapes colliding at the position
      :rtype: list(Shape)

      Get a list of all shapes colliding with shape
      """
      collisions = []
      id = self.ids[shape]
      for other in self.grid.get(self.areas[shape]):
         if self.collidable[other]:
            if id != self.ids[other]:
               if self.collider.rect_rect(shape.width, shape.height, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1]):
                  collisions.append(other)
      return collisions

   def collisions_with_group(self, group):
      collisions = []
      group = self.groups[group]
      for shape in group:
         if self.collidable[shape]:
            id = self.ids[shape]
            for other in self.grid.get(self.areas[shape]):
               if self.collidable[other]:
                  if id != self.ids[other]:
                     if other not in group:
                        if self.collider.rect_rect(shape.width, shape.height, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1]):
                           collisions.append(other)
      return collisions

   def check_at(self, x, y, width=1, height=1):
      """
      :param x: Posistion to search for on the x-axis
      :param y: Posistion to search for on the y-axis
      :param width: Width of the search area
      :param height: Height of the search area
      :type x: Int
      :type y: Int
      :type width: Int
      :type height: Int
      :return: True if there is a collision at x, y, False if not
      :rtype: bool

      Check to see if there is a collision at x, y in space.
      """
      for other in self.grid.get(self.grid.scale(x, y)):
         if self.collidable[other]:
            if self.collider.rect_rect(width, height, x, y, other.width, other.height, other.position[0], other.position[1]):
               return True
      return False
      
   def check(self, shape):
      """
      :param shape: Shape to check for collisions
      :type shape: Shape
      :return: True if the shape is colliding with anything in space, False if not
      :rtype: bool

      Check to see if something is colliding with shape in the space.
      """
      id = self.ids[shape]
      for other in self.grid.get(self.areas[shape]):
         if self.collidable[other]:
            if id != self.ids[other]:
               if self.collider.rect_rect(shape.width, shape.height, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1]):
                  return True
      return False

   def check_group(self, group):
      group = self.groups[group]
      for shape in group:
         if self.collidable[shape]:
            id = self.ids[shape]
            for other in self.grid.get(self.areas[shape]):
               if self.collidable[other]:
                  if id != self.ids[other]:
                     if other not in group:
                        if self.collider.rect_rect(shape.width, shape.height, shape.position[0], shape.position[1], other.width, other.height, other.position[0], other.position[1]):
                           return True
      return False


   def check_two(self, shape, other):
      """
      :param shape: First Shape to check for collision
      :param other: Second Shape to check for collision
      :type shape: Shape
      :type other: Shape
      :return: True if the two shapes are colliding, False if not
      :rtype: Bool

      Check to see if two shapes are colliding
      """
      return self.collider.check_rects(shape, other)

   def update_area(self, shape):
      """
      :param shape: Body to be moved from the space
      :param area: Area of the space to move the shape to
      :type shape: AbsBody
      :type area: Tuple(Int, Int)
      :return: Original area if updating, None if not
      :rtype: Tuple(Int, Int) or None

      Moves a shape from one collision area to another
      """
      area = self.grid.scale(shape.position[0], shape.position[1])
      start = self.areas[shape]
      if start != area:
         self.move_area(shape, area)
         return start
      else:
         return None

   def move_area(self, shape, area):
      #TODO this needs to be optimized better but I'm not doing it right now
      start = self.areas[shape]
      self.delete_from_grid(shape, area)
      self.add_to_grid(shape, area)
      self.areas[shape] = area

   def delete_from_grid(self, shape, area):
      self.grid.delete(shape, area)
      self.grid.delete(shape, (area[0]+1, area[1]))
      self.grid.delete(shape, (area[0], area[1]+1))
      self.grid.delete(shape, (area[0]+1, area[1]+1))

   def add_to_grid(self, shape, area):
      self.grid.add(shape, area)
      self.grid.add(shape, (area[0]+1, area[1]))
      self.grid.add(shape, (area[0], area[1]+1))
      self.grid.add(shape, (area[0]+1, area[1]+1))
      
   def move(self, shape, distance):
      """
      :param shape: Shape to move
      :param x: Direction to move in on horizontal plane
      :param y: Direction to move in on the vertical plane
      :param distance: Distance to move for
      :type shape: Shape
      :type x: int
      :type y: iny
      :type distance: float
      :return: Distance moved
      :rtype: tuple(int, int)
      
      Advances a shape in its direction for distance units.
      """
      direction = self.directions[shape]
      xstep = int(direction[0] * distance)
      ystep = int(direction[1] * distance)
      shape.position = (shape.position[0]+xstep, shape.position[1]+ystep)
      self.update_area(shape)
      return (xstep, ystep)

   def move_group(self, group, distance):
      #TODO this could be optimized in some/most cases by uniform group movement to not recalc the direction a shape moves for every shape.
      for shape in self.groups[group]:
         self.move(shape, distance)

   def push(self, shape, x, y):
      """
      :param shape: Shape to push
      :param x: Direction on the horizontal plane to push the shape
      :param y: Direction on the vertical plane to push the shape
      :type shape: Shape
      :type x: int
      :type y: int

      Pushes a shape in the direction of x, y in space
      """
      shape.position = (shape.position[0]+x, shape.position[1]+y)
      self.update_area(shape)

   def push_group(self, group, x, y):
      for shape in self.groups[group]:
         self.push(shape, x, y)

   def place(self, shape, x, y):
      """
      :param x: The position in space to place the shape on the horizontal plane
      :param y: The position in space to place the shape on the vertical plane
      :type x: int
      :type y: int

      Directly places a shape at x, y in space
      """
      shape.position = (x, y)
      self.update_area(shape)

   def place_group(self, group, x, y):
      for shape in self.groups[group]:
         self.place(shape, x, y)


   def set_active(self, shape, value):
      if value:
         if shape not in self.actives:
            self.actives.append(shape)
      else:
         if shape in self.actives:
            self.actives.remove(shape)

   def set_active_group(self, group, value):
      for shape in self.groups[group]:
         self.set_active(shape, value)

   def set_moving(self, shape, value):
      if value:
         if shape not in self.moving:
            self.moving.append(shape)
      else:
         if shape in self.moving:
            self.moving.remove(shape)

   def set_moving_group(self, group, value):
      for shape in self.groups[group]:
         self.set_moving(shape, value)

   def set_collidable(self, shape, value):
      self.collidable[shape] = value

   def set_collidable_group(self, group, value):
      for shape in self.groups[group]:
         self.collidable[shape]= value
         
   def turn(self, shape, direction):
      self.directions[shape] = direction

   def turn_group(self, group, direction):
      for shape in self.groups[group]:
         self.directions[shape] = direction

   def step(self, delta):
      """
      :param delta: Units of time to advance the simulation
      :type delta: float
      Process the system for delta steps.
      Triggers the on_step_start(delta), on_step(body, delta), on_collision(body, collisions), and on_step_end(delta) callbacks
      """
      self.on_step_start(delta)
      for shape in self.moving:
         self.on_step(shape, delta)
         self.move(shape, delta)

      #TODO This is very unoptimized
      for shape in self.actives:
         if self.collidable[shape]:
            collisions = self.collisions_with(shape)
            if collisions:
               self.on_collision(shape, collisions)
      self.on_step_end(delta)

   def on_step_start(self, delta):
      """
      :param delta: Units of time to advance the simulation
      :type delta: float
      Callback before the simulation takes a step. Override this when extending your Controller
      """
      return


   def on_step(self, shape, delta):
      """
      :param body: Body being processed by the step
      :param delta: Units of time to advance the simulation
      :type body: Body
      :type delta: float
      Callback when the simulation takes a step over a body. Override this when extending your Controller
      """
      return

   def on_collision(self, shape, collisions):
      """
      :param body: Body involved in a collision
      :param collision: List of bodies colliding with body
      :type body: Body
      :type other: List(Body)
      Callback when bodies collide with another body. Override this when extending your Controller
      """
      return

   def on_step_end(self, delta):
      """
      :param delta: Units of time to advance the simulation
      :type delta: float
      Callback after the simulation takes a step. Override this when extending your Controller
      """
      return
      



from .shape import Shape
from .shape import Rect
from .shape import Circle

from .clock import Clock
from .clock import SyncClock

from .index import Index
from .spatialhash import SpatialHash
from .collider import Collider
from .space import Space

__all__ = ["Shape", "Rect", "Circle", "Clock", "SyncClock", "Index", "SpatialHash", "Collider", "Space"]
version = 0.035


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pecrs']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pecrs',
    'version': '0.35',
    'description': 'Pythonic Entity Collision Resolution System',
    'long_description': '![logo](https://raw.githubusercontent.com/solidsmokesoftware/pecrs-py/master/logo.png)\n\n# Pythonic Entity Collision Resolution System\n\npecrs is a pure Python 2D physics system with a focus on top-down games and simple platformers. \n\nPure Python makes pecrs portable and easy to modify to suit your own needs.\n\nFocused use-case makes pecrs simple to learn and use.\n\n[Seamless integration](https://solidsmokesoftware.github.io/pecrs-py/pyglet.html) with [Pyglet](http://pyglet.org/)\n\n# Installation\n\nVia pip\n\n`python3 -m pip install pecrs`\n\n# Quickstart\n```python\n\nfrom pecrs import *\n\nspace = Space()\nrectA = Rect(0, 0, 32, 32)\nrectB = Rect(10, 0, 32, 32)\n\nspace.add(rectA)\nspace.add(rectB)\n\ncollision = space.check(rectA)\nprint(f"Is something colliding with rectA? {collision}")\n\ncollisions = space.collisions_with(rectB)\nprint(f"Who is colliding with rectB? {collisions}")\n\nspace.place(rectB, 100, 0)\n\ncollision = space.check_two(rectA, rectB)\nprint(f"Are rectA and rectB still colliding? {collision}")\n```\n\n# Structual Overview\n\nBase types of the system are Shapes and Bodies. \nShapes have a position and dimensions which describe its physical properties.\nBodies are Shapes with an id, direction, speed, and movement state.\n\nCore functionality is providied by the Collider, which detects collisions between Shapes or Shape-like Objects.\n\nThe Space handles positioning of Shapes and optimizes collision handling.\n\nThe Controller provides high-level object oriented control over Bodies in a Space.\n\n# Real-world Usage\n\n![demo](https://raw.githubusercontent.com/solidsmokesoftware/pecrs-py/master/pyglet_demo.gif)\n\n```python\n\nfrom pecrs import *\nimport pyglet\n\n\nclass World(Space):\n   def __init__(self):\n      super().__init__()\n      self.window = pyglet.window.Window(400, 300)\n      self.batch = pyglet.graphics.Batch()\n      \n      self.red_image = pyglet.resource.image("red_rect.png")\n      self.blue_image = pyglet.resource.image("blue_rect.png")\n\n      spriteA = pyglet.sprite.Sprite(self.blue_image, x=0, y=150, batch=self.batch)\n      spriteB = pyglet.sprite.Sprite(self.blue_image, x=300, y=150, batch=self.batch)\n\n      self.add(spriteA, moving=True)\n      self.turn(spriteA, (150, 0))\n\n      self.add(spriteB)\n      \n      pyglet.clock.schedule_interval(self.run, 1.0/60)\n      pyglet.app.run()\n\n   def on_collision(self, shape, collisions):\n      shape.image = self.red_image\n\n   def run(self, delta):\n      self.step(delta)\n      self.window.clear()\n      self.batch.draw()\n\nworld = World()\n```\n\n# Documentation\n\nhttps://solidsmokesoftware.github.io/pecrs-py/\n\n# Demonstration\n\nhttps://github.com/solidsmokesoftware/solconomy\n\n![solconomy](https://camo.githubusercontent.com/de20b3b2014d20a8746f7346e777e323586d5a35/68747470733a2f2f692e696d6775722e636f6d2f566277677664372e706e67)\n\n# Requirements\n\nTested with Python3.6.9\n',
    'author': 'Solid Smoke Software',
    'author_email': 'solid.smoke.software@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/solidsmokesoftware/pecrs-py',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.9,<4.0.0',
}


setup(**setup_kwargs)
